package gregapi.tileentity.energy;

import static gregapi.data.CS.*;
import gregapi.code.ArrayListNoNulls;
import gregapi.code.TagData;
import gregapi.data.LH;
import gregapi.data.TD;
import gregapi.tileentity.base.TileEntityBase09FacingSingle;
import gregapi.tileentity.machines.ITileEntityRunningActively;
import gregapi.util.UT;

import java.util.Collection;
import java.util.List;

import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;

/**
 * @author Gregorius Techneticies
 */
public abstract class TileEntityBase10EnergyConverter extends TileEntityBase09FacingSingle implements ITileEntityEnergy, ITileEntityRunningActively {
	protected boolean mWasteEnergy = F, mEmitsEnergy = F, mStopped = F, mActive = F;
	protected long mEnergy = 0, mInput = 32, mOutput = 16, mMultiplier = 1, mActiveData = 0;
	protected byte mActiveState = 0, mExplosionPrevention = 0;
	protected TagData mEnergyTypeAccepted = TD.Energy.QU;
	protected TagData mEnergyTypeEmitted = TD.Energy.QU;
	
    @Override
	public void readFromNBT2(NBTTagCompound aNBT) {
    	super.readFromNBT2(aNBT);
    	mEnergy = aNBT.getLong(NBT_ENERGY);
        if (aNBT.hasKey(NBT_ACTIVE_ENERGY)) mEmitsEnergy = aNBT.getBoolean(NBT_ACTIVE_ENERGY);
        if (aNBT.hasKey(NBT_WASTE_ENERGY)) mWasteEnergy = aNBT.getBoolean(NBT_WASTE_ENERGY);
        if (aNBT.hasKey(NBT_STOPPED)) mStopped = aNBT.getBoolean(NBT_STOPPED);
        if (aNBT.hasKey(NBT_ACTIVE)) mActive = aNBT.getBoolean(NBT_ACTIVE);
        if (aNBT.hasKey(NBT_ACTIVE_DATA)) {mActiveData = aNBT.getLong(NBT_ACTIVE_DATA);}
        if (aNBT.hasKey(NBT_INPUT)) {mInput = aNBT.getLong(NBT_INPUT);}
        if (aNBT.hasKey(NBT_OUTPUT)) {mOutput = aNBT.getLong(NBT_OUTPUT);}
        if (aNBT.hasKey(NBT_MULTIPLIER)) {mMultiplier = aNBT.getLong(NBT_MULTIPLIER);}
        if (aNBT.hasKey(NBT_ENERGY_EMITTED)) mEnergyTypeEmitted = TagData.createTagData(aNBT.getString(NBT_ENERGY_EMITTED));
        if (aNBT.hasKey(NBT_ENERGY_ACCEPTED)) mEnergyTypeAccepted = TagData.createTagData(aNBT.getString(NBT_ENERGY_ACCEPTED));
    }
    
    @Override
	public void writeToNBT2(NBTTagCompound aNBT) {
    	super.writeToNBT2(aNBT);
    	UT.NBT.setNumber(aNBT, NBT_ACTIVE_DATA, mActiveData);
    	UT.NBT.setNumber(aNBT, NBT_ENERGY, mEnergy);
        UT.NBT.setBoolean(aNBT, NBT_ACTIVE, mActive);
        UT.NBT.setBoolean(aNBT, NBT_STOPPED, mStopped);
        UT.NBT.setBoolean(aNBT, NBT_ACTIVE_ENERGY, mEmitsEnergy);
    }
    
	@Override
	public void addToolTips(List aList, ItemStack aStack, boolean aF3_H) {
		addToolTipsEnergy(aList, aStack, aF3_H);
		addToolTipsEfficiency(aList, aStack, aF3_H);
		super.addToolTips(aList, aStack, aF3_H);
	}
	
	public void addToolTipsEnergy(List aList, ItemStack aStack, boolean aF3_H) {
		LH.addEnergyToolTips(this, aList, mEnergyTypeAccepted, mEnergyTypeEmitted, getLocalisedInputSide(), getLocalisedOutputSide());
	}
	
	public void addToolTipsEfficiency(List aList, ItemStack aStack, boolean aF3_H) {
		if (TD.Energy.ALL_EU.contains(mEnergyTypeAccepted)) {
			if (TD.Energy.ALL_EU.contains(mEnergyTypeEmitted)) {
				aList.add(LH.getToolTipEfficiency(UT.Code.units(10000, mInput, mOutput*mMultiplier, F)));
			} else {
				if (mEnergyTypeEmitted == TD.Energy.RF		) aList.add(LH.getToolTipEfficiency(UT.Code.units(10000, mInput*RF_PER_EU, mOutput*mMultiplier, F)));
			}
		} else {
			if (TD.Energy.ALL_EU.contains(mEnergyTypeEmitted)) {
				if (mEnergyTypeAccepted == TD.Energy.RF		) aList.add(LH.getToolTipEfficiency(UT.Code.units(10000, mInput, mOutput*mMultiplier*RF_PER_EU, F)));
				if (mEnergyTypeAccepted == TD.Energy.STEAM	) aList.add(LH.getToolTipEfficiency(UT.Code.units(10000, mInput, mOutput*mMultiplier*STEAM_PER_EU, F)));
			}
		}
	}
	
    @Override
    public void onTick2(long aTimer, boolean aIsServerSide) {
    	if (aIsServerSide) {
    		doConversion(aTimer);
			if (mTimer % 600 == 5) if (mActive) doDefaultStructuralChecks(); else if (mExplosionPrevention > 0) mExplosionPrevention--;
    	}
    }
    
    @Override
    public boolean onTickCheck(long aTimer) {
    	mActiveData <<= 1;
    	if (mActive) mActiveData |= 1;
    	byte tActiveState = mActiveState;
    	if (mActiveData == 0 || mStopped) mActiveState = 0; else if (mActiveData == ~0L) mActiveState = 1; else mActiveState = 2;
    	return tActiveState != mActiveState || super.onTickCheck(aTimer);
    }
    
	@Override
	public void setVisualData(byte aData) {
		mActiveState = aData;
	}
	
	@Override
	public long doInject(TagData aEnergyType, byte aSide, long aSize, long aAmount, boolean aDoInject) {
		aSize = Math.abs(aSize);
		if (aSize > getEnergySizeInputMax(aEnergyType, aSide)) {
			if (aDoInject) overload(aSize, aEnergyType);
			return aAmount;
		}
		long tMaxBuffer = getMaxBuffered(aEnergyType, aSide);
		if (mEnergy >= tMaxBuffer) return 0;
		long tInput = Math.min(tMaxBuffer - mEnergy, aSize * aAmount), tConsumed = Math.min(aAmount, (tInput/aSize) + (tInput%aSize!=0?1:0));
		if (aDoInject) mEnergy += tConsumed * aSize;
		return tConsumed;
	}
	
	public void overload(long aSize, TagData aEnergyType) {
		if (mExplosionPrevention < 100) {
			if (mTimer < 100) DEB.println("Machine overloaded on startup with: " + aSize + " " + aEnergyType.getLocalisedNameLong());
			mExplosionPrevention++;
			mEnergy = 0;
		} else {
			overcharge(aSize, aEnergyType);
		}
	}
	
	public long getMaxBuffered(TagData aEnergyType, byte aSide) {return getEnergySizeInputMax(aEnergyType, aSide);}
	
	@Override public boolean isEnergyType					(TagData aEnergyType, byte aSide, boolean aEmitting) {return aEmitting ? aEnergyType == mEnergyTypeEmitted : aEnergyType == mEnergyTypeAccepted;}
	@Override public boolean isEnergyAcceptingFrom			(TagData aEnergyType, byte aSide, boolean aTheoretical) {return (aTheoretical || !mStopped) && (SIDES_INVALID[aSide] || isInput (aSide)) && super.isEnergyAcceptingFrom(aEnergyType, aSide, aTheoretical);}
	@Override public boolean isEnergyEmittingTo				(TagData aEnergyType, byte aSide, boolean aTheoretical) {return                                (SIDES_INVALID[aSide] || isOutput(aSide)) && super.isEnergyEmittingTo   (aEnergyType, aSide, aTheoretical);}
	@Override public long getEnergySizeOutputRecommended	(TagData aEnergyType, byte aSide) {return mOutput;}
	@Override public long getEnergySizeInputRecommended		(TagData aEnergyType, byte aSide) {return mInput;}
	@Override public Collection<TagData> getEnergyTypes(byte aSide) {return new ArrayListNoNulls<TagData>(F, mEnergyTypeAccepted, mEnergyTypeEmitted);}
	
	// Stuff to Override
	
	public void doConversion(long aTimer) {
		long tOutput = UT.Code.units(mEnergy, mInput, mOutput, F), tMaxOut = getEnergySizeOutputMax(mEnergyTypeEmitted, SIDE_ANY);
		mActive = (tOutput >= getEnergySizeOutputMin(mEnergyTypeEmitted, SIDE_ANY));
		mEmitsEnergy = F;
		
		if (mActive) {
			if (tOutput > tMaxOut && TD.Energy.ALL_COMSUMPTION_LIMITED.contains(mEnergyTypeAccepted)) tOutput = tMaxOut;
			if (tOutput > tMaxOut) {
				if (mTimer > 2) {
					overload(mEnergy, mEnergyTypeEmitted);
				} else {
					mEnergy = 0;
					DEB.println("Machine overcharged on startup with: " + mEnergy + " " + mEnergyTypeEmitted.getLocalisedNameLong());
				}
			} else {
				if (TD.Energy.ALL_SIZE_IRRELEVANT.contains(mEnergyTypeEmitted)) {
					long tEmittedPackets = ITileEntityEnergy.Util.emitEnergyToNetwork(mEnergyTypeEmitted, 1, tOutput*mMultiplier, this);
					if (!mWasteEnergy) mEnergy -= UT.Code.units(tEmittedPackets, mOutput*mMultiplier, mInput, T);
					mEmitsEnergy = (tEmittedPackets > 0);
				} else {
					long tEmittedPackets = ITileEntityEnergy.Util.emitEnergyToNetwork(mEnergyTypeEmitted, tOutput, mMultiplier, this);
					if (!mWasteEnergy) mEnergy -= UT.Code.units(tEmittedPackets * tOutput, mOutput*mMultiplier, mInput, T);
					mEmitsEnergy = (tEmittedPackets > 0);
				}
			}
		}
		if (mWasteEnergy) mEnergy = Math.max(0, mEnergy-getEnergySizeInputMax(mEnergyTypeAccepted, SIDE_ANY));
	}
	
	@Override public boolean getStateRunningPossible() {return T;}
	@Override public boolean getStateRunningPassively() {return mActive;}
	@Override public boolean getStateRunningActively() {return mEmitsEnergy;}
	public boolean setStateOnOff(boolean aOnOff) {mStopped = !aOnOff; return !mStopped;}
	public boolean getStateOnOff() {return !mStopped;}
	
	@Override public byte getVisualData() {return mActiveState;}
	
	public boolean isInput (byte aSide) {return aSide != mFacing;}
	public boolean isOutput(byte aSide) {return aSide == mFacing;}
	public String getLocalisedInputSide () {return LH.get(LH.FACE_ANYBUT_FRONT);}
	public String getLocalisedOutputSide() {return LH.get(LH.FACE_FRONT);}
}